{ File   : CMP2BMP.dpr
  Author : Deniz Oezmen
  Created: 2005-08-20
  Changed: 2005-10-08

  Star Trek: DS9 Harbinger CMP background decompressor.
}
program CMP2BMP;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils, SaveBMP2, DePKDCL;

type
  TCMPHeader    = packed record
                    DataSize: Longword;
                  end;

  TBuffer       = array of Byte;

var
  CMPFile  : file;
  CMPHeader: TCMPHeader;
  Palette  : TPalette3;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenCMPFile(var CMPFile: file; CMPName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(CMPFile, CMPName);

  {$I-}
  Reset(CMPFile, 1);
  {$I+}
  OpenCMPFile := (IOResult = 0)
end;

function ReadHeader(var CMPHeader: TCMPHeader): Boolean;
var
  BytesRead: Longint;
begin
  BlockRead(CMPFile, CMPHeader, SizeOf(CMPHeader), BytesRead);
  ReadHeader := (BytesRead = SizeOf(CMPHeader)) and
    (Longword(FileSize(CMPFile)) >= CMPHeader.DataSize + 4)
end;

procedure ProcessCMP(CMPName: string; CMPHeader: TCMPHeader);
var
  InBuffer,
  OutBuffer: TDCLBuffer;
begin
  Write(CMPName, ' => ');

  CMPName := Prefix(CMPName) + '.bmp';

  WriteLn(CMPName);

  SetLength(InBuffer, CMPHeader.DataSize);
  SetLength(OutBuffer, 0);

  BlockRead(CMPFile, InBuffer[0], CMPHeader.DataSize);
  Blast(InBuffer, OutBuffer);

  Dump8BitBMP(CMPName, 640, 480, Palette, OutBuffer)
end;

function ReadPalette(PalName: string): Boolean;
var
  i, t   : Byte;
  PalFile: file;
begin
  FileMode := fmOpenRead;
  AssignFile(PalFile, PalName);

  {$I-}
  Reset(PalFile, 1);
  {$I+}
  Result := (IOResult = 0) and (FileSize(PalFile) = 768);

  if not Result then
    Exit;

  BlockRead(PalFile, Palette, SizeOf(Palette));

  for i := 0 to 255 do
    with Palette[i] do
    begin
      t := R;
      R := B shl 2;
      G := G shl 2;
      B := t shl 2
    end
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  CMPName   : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara 
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara 
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      CMPName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenCMPFile(CMPFile, CMPName) then
        if ReadHeader(CMPHeader) then
        begin
          ProcessCMP(CMPName, CMPHeader);
          Close(CMPFile)
        end
        else
          Writeln(CMPName, ' is not a valid CMP file.')
      else
        Writeln(CMPName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found');
  FreeMem(PFilePara, Length(FilePara) + 1)
end;

begin
  if ParamCount = 2 then
  begin
    if ReadPalette(ParamStr(2)) then
    begin
      InitBlast;
      ParseFiles(ParamStr(1))
    end
    else
      WriteLn('Palette file not accessible or invalid.')
  end
  else
    WriteLn('Incorrect number of parameters.')
end.
