{ File   : RLE2BMP.dpr
  Author : Deniz Oezmen
  Created: 2005-08-17
  Changed: 2005-10-08

  Star Trek: DS9 Harbinger RLE decompressor.
}
program RLE2BMP;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils, SaveBMP2;

type
  TRLEHeader    = packed record
                    Unknown  : array[0..7] of Byte;
                    NrSprites: Word
                  end;

  TSpriteHeader = packed record
                    Unknown1: Byte;
                    Width,
                    Height  : Word;
                    Unknown2: array[0..15] of Byte;
                    DataSize: Longword
                  end;

  TBuffer       = array of Byte;

var
  RLEFile  : file;
  RLEHeader: TRLEHeader;
  Palette  : TPalette3;

function IntToStrL(i: Longint; Len: Byte): string;
begin
  Result := IntToStr(i);
  while Length(Result) < Len do
    Result := '0' + Result
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenRLEFile(var RLEFile: file; RLEName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(RLEFile, RLEName);

  {$I-}
  Reset(RLEFile, 1);
  {$I+}
  OpenRLEFile := (IOResult = 0)
end;

function ReadHeader(var RLEHeader: TRLEHeader): Boolean;
var
  BytesRead: Longint;
begin
  BlockRead(RLEFile, RLEHeader, SizeOf(RLEHeader), BytesRead);
  ReadHeader := (BytesRead = SizeOf(RLEHeader))
end;

procedure RLEDecode(InBuffer, OutBuffer: TBuffer; InSize, OutSize: Longint;
  Width: Word);
var
  InPos,
  OutPos: Longint;
  i,
  b, c  : Byte;
begin
  InPos := 0;
  OutPos := 0;

  while InPos <= InSize - 1 do
  begin
    b := InBuffer[InPos];
    Inc(InPos);

    if b and $40 > 0 then
    begin
      c := b shr 7;
      b := b and $3f;

      if c = 0 then
      begin
        while OutPos mod Width > 0 do
        begin
          OutBuffer[OutPos] := c;
          if OutPos = OutSize - 1 then
            Break;
          Inc(OutPos)
        end
      end;

      for i := 1 to b do
      begin
        OutBuffer[OutPos] := c;
        if OutPos = OutSize - 1 then
          Break;
        Inc(OutPos)
      end
    end
    else if b and $80 > 0 then
    begin
      b := b and $7f;
      c := InBuffer[InPos];
      Inc(InPos);
      for i := 0 to b - 1 do
      begin
        OutBuffer[OutPos] := c;
        if OutPos = OutSize - 1 then
          Break;
        Inc(OutPos)
      end
    end
    else
    begin
      for i := 1 to b do
      begin
        OutBuffer[OutPos] := InBuffer[InPos];
        Inc(InPos);
        if OutPos = OutSize - 1 then
          Break;
        Inc(OutPos)
      end
    end
  end
end;

procedure ProcessRLE(RLEName: string; RLEHeader: TRLEHeader);
var
  i           : Word;
  InSize,
  OutSize     : Longword;
  SpriteHeader: TSpriteHeader;
  InBuffer,
  OutBuffer   : TBuffer;
begin
  WriteLn(RLEName, ': ', RLEHeader.NrSprites, ' sprites');

  RLEName := Prefix(RLEName);

  for i := 1 to RLEHeader.NrSprites do
  begin
    BlockRead(RLEFile, SpriteHeader, SizeOf(SpriteHeader));

    InSize := SpriteHeader.DataSize;
    OutSize := SpriteHeader.Width * SpriteHeader.Height;

    SetLength(InBuffer, InSize);
    SetLength(OutBuffer, OutSize);

    BlockRead(RLEFile, InBuffer[0], InSize);
    RLEDecode(InBuffer, OutBuffer, InSize, OutSize, SpriteHeader.Width);

    WriteLn(' => ', RLEName + '#' + IntToStrL(i, 5) + '.bmp');

    Dump8BitBMP(RLEName + '#' + IntToStrL(i, 5) + '.bmp', SpriteHeader.Width,
      SpriteHeader.Height, Palette, OutBuffer)
  end
end;

function ReadPalette(PalName: string): Boolean;
var
  i, t   : Byte;
  PalFile: file;
begin
  FileMode := fmOpenRead;
  AssignFile(PalFile, PalName);

  {$I-}
  Reset(PalFile, 1);
  {$I+}
  Result := (IOResult = 0) and (FileSize(PalFile) = 768);

  if not Result then
    Exit;

  BlockRead(PalFile, Palette, SizeOf(Palette));

  for i := 0 to 255 do
    with Palette[i] do
    begin
      t := R;
      R := B shl 2;
      G := G shl 2;
      B := t shl 2
    end
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  RLEName   : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara 
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara 
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      RLEName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenRLEFile(RLEFile, RLEName) then
        if ReadHeader(RLEHeader) then
        begin
          ProcessRLE(RLEName, RLEHeader);
          Close(RLEFile)
        end
        else
          Writeln(RLEName, ' is not a valid RLE file.')
      else
        Writeln(RLEName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found');
  FreeMem(PFilePara, Length(FilePara) + 1)
end;

begin
  if ParamCount = 2 then
  begin
    if ReadPalette(ParamStr(2)) then
      ParseFiles(ParamStr(1))
    else
      WriteLn('Palette file not accessible or invalid.')
  end
  else
    WriteLn('Incorrect number of parameters.')
end.
